# CMake based Conveyor template

This template project sets up a CMake based C++ project with Conveyor packaging. It can be compiled for Windows, Mac ARM/x86 and Linux.

The app demonstrates the following useful patterns:

1. Setting up a CMake build that can produce executables on each operating system that Conveyor supports.
2. Depending on a third party library with automatic downloading and compilation.
3. Using dynamically linked libraries and setting the dynamic linker search path appropriately.
4. Creating fat binaries on macOS (fat binaries contain both x86 and ARM code in one file). 
5. Packaging the resulting native app with Conveyor.

## OpenGL

This template is meant to be a generic starting point for your own programs. See below for how to customize it for your own needs.

To ensure the template has something to do it sets up and uses an OpenGL context with [a beautiful pixel shader by the talented
bitless](https://www.shadertoy.com/user/bitless), using the cross-platform GLFW OpenGL library. GLFW handles the differences between each
operating system's windowing APIs and thus keeps the sample code simple and portable. If you don't want to use OpenGL, obviously just delete
everything related to that.

## How to build and package

### Outline

Creating a download site with packages involves the following steps:

1. Compile the app on each OS you wish to support.
2. Either transfer the compiled files to a single machine and ensure they're in the right directories, or, expose them via HTTP and put
   the URLs in your `conveyor.conf` file.
3. Run `conveyor make site`.

### Step 1. Choose your supported platforms.

The `conveyor.conf` file expects the binaries for each platform to appear in the following directories relative to the project root:

* `build/windows/installation` for Windows
* `build/mac/installation` for Mac
* `build/linux/installation` for Linux 

The template doesn't set up cross-compilation so the executables for each operating system must be compiled on that operating system. You
might not have access to all three of these platforms. If not, pick the ones you want and then uncomment the `machines = [ .. ]` line in
the `conveyor.conf` file. It should contain a comma separated list of platforms. For example to exclude macOS use `machines = [ windows, linux ]`.
You could also specify the CPU architecture, for example `machines = [ windows.amd64, linux.aarch64 ]`.

Each OS will require both [CMake](https://cmake.org/) and the appropriate developer tools installed:

* **Windows**: The template has been tested with Visual Studio 2019 but other releases will probably work.
* **macOS**: You'll need to get Xcode from the App Store.
* **Linux**: Install your distribution's compiler toolchain. You may also need to install extra development libraries. On Ubuntu/Debian, try `sudo apt install libx11-dev xserver-xorg-dev xorg-dev` first.

### Step 2. Compile

On each machine create the appropriate directory _without_ the `installation` suffix and then invoke CMake to create the build system and 
run the "install" target. This will create a runnable directory structure with `bin` and `lib` directories under the build directory.

=== "Windows"

    ```
    mkdir build\windows
    cd build\windows
    cmake ..\..
    cmake --build . --config Release --target install
    ```

=== "macOS"

    ```bash
    mkdir -p build/mac
    cd build/mac
    cmake ../..
    make install
    ```

=== "Linux"

    ```bash
    mkdir -p build/linux
    cd build/linux
    cmake ../..
    make install
    ```

You should now have a runnable program in `build\windows\installation\bin` or `build/{mac,linux}/installation/bin`.

### Step 3. Bring the compiled binaries together

Conveyor needs the binaries for each platform. You can do this in two different ways:

1. Copy the `installation` directories from each OS and place them into the project directory hierarchy.
2. Replace the file paths in the `conveyor.conf` file with URLs to where the binaries can be downloaded. See the Conveyor user guide for 
   how to use URLs and archives as inputs.

A convenient way to get the files in the right place is by compiling inside virtual machines that share a disk. If copying the files into
the project tree, you should now have a directory structure like this:

* Project root
  * `CMakeLists.txt`
  * `conveyor.conf`
  * `src/*`
  * `build`
    * `windows`
      * `installation`
        * `bin/gl_cmake.exe`
        * `bin/glfw3.dll`
        * `lib/glfw3.lib`
    * `mac`
      * `installation`
        * `bin/gl_cmake`
        * `lib/libglfw.dylib` (and symlinks)
    * `linux`
      * `installation`
        * `bin/gl_cmake`
        * `lib/libglfw.so` (and symlinks)

### Step 4. Build the download site

Now simply run `conveyor make site`. You should get a directory called `output` with the packages for each platform.

## Adapting the template

To strip the app down to the basics:

* Delete the contents of the `include/` and `src/` directories.
* Replace `src/main.cpp` with a hello world style `int main(int argc, char **argv) {}` entry point.
* Delete the `ImportGLFW.txt` file and the two lines in `CMakeLists.txt` that include it and add a library dependency on glfw.

This will remove all the OpenGL specific stuff, leaving you with a generic portable build system.

## License

This template is made available under the Apache 2 License, with the exception of the pixel shader which is under the 
[Creative Commons Attribution-NonCommercial-ShareAlike 3.0 Unported License](https://creativecommons.org/licenses/by-nc-sa/3.0/deed.en_US).
